@@ -1,121 +1,32 @@
 /*
-
-   /dev/3dfx device for 2.x kernels with MTRR settings enabled.
-
-   Compile with :
-
-   gcc -O2 -DMODULE -D__KERNEL__ -I/usr/src/linux/include -pipe
-       -fno-strength-reduce -malign-loops=2 -malign-jumps=2
-       -malign-functions=2 -c -o 3dfx.o 3dfx_driver.c
-
-   Add -D__SMP__ if you're running of an multi-processor system.
-   Add -DHAVE_MTRR if you have and want MTRR settings.
-   Add -g -Wall -Wstrict-prototypes -DDEBUG, if you are debugging.
-
-   Original by Daryll Straus.
-   Port to 2.1 kernel by Jon Taylor.
-   setmtrr_3dfx() added by Jens Axboe.
-   Combining 2.0 and 2.1/2.2 kernels into one rpm,
-     resetmtrr_3df(), and using correct pci calls for
-     2.1/2.2 kernels by Carlo Wood.
-
-   ChangeLog
-
-   2004-05-05   Guillem Jover   <guillem@debian.org>
-
-   * Use designators in fops_3dfx instead of the obsolete colon syntax.
-
-   2004-05-04   Guillem Jover   <guillem@debian.org>
-
-   * Cleaned new kernel build system.
-   * Indentation fixes. Normalized ChangeLog date format.
-
-   2004-02-13   Aristeu Sergio Rozanski Filho <aris@cathedrallabs.org>
-
-   * Use kernel build system.
-   * Use new pci probe present in 2.6.x kernels.
-   * Use module auto-loading.
-
-   2004-01-22   Guillem Jover   <guillem@debian.org>
-
-   * Ported to 2.6 kernels.
-
-   2002-05-05   Guillem Jover   <guillem@hadrons.org>
-
-   * Re-included VOODOO pci id, accidentally removed in last patch,
-     thanks to Jan Kuipers to point this out.
-
-   2002-01-29   Guillem Jover   <guillem@hadrons.org>
-
-   * Fixed bug when compiling for 2.4 kernels with modversions.
-   * Added devfs support.
-   * Added MODULE_* declarations.
-   * Cleaned detection code.
-
-   2000-02-04   Joseph Kain     <joseph@3dfx.com>
-
-   * Updated Carlo Woord's email address.
-
-   1999-12-22   Joseph Kain     <joseph@3dfx.com>
-
-   * Support for new VMA structure in Kernels 2.3.14 and higher.
-     This patch is based on Dieter Nuetzel's work.
-
-   1999-11-09   Joseph Kain     <joseph@3dfx.com>
-
-   * Made cards static to remove conflict with hisax ISDN driver.
-
-   1998-10-30   Carlo Wood      <carlo@alinoe.com>
-
-   * Re-included module versions (this source was based on the
-     source by Jon Taylor which was based on an older version
-     by Daryll Straus.  I now included the later changes made
-     by Daryll in this source too).
-   * Use pci_find_device for 2.1/2.2 kernels instead of the old
-     pcibios_find_device.  Idem pci_present().
-
-   1998-10-21   Carlo Wood      <carlo@alinoe.com>
-
-   * Fixed compile bugs related to 2.0 kernels.
-
-   1998-10-21   Carlo Wood      <carlo@alinoe.com>
-
-   * `struct file_operations' was changed as of kernel 2.1.118,
-     fixed this module to also work with versions > 2.1.117.
-
-   * Machines that don't have mtrr still need <asm/uaccess.h>.
-
-   1998-10-20   Carlo Wood      <carlo@alinoe.com>
-
-   * Finally fixed in init_module() the right way.  This time I tested
-     it too (it now really works).
-
-   * Added resetmtrr_3dfx() to clean up the MTRR settings when the
-     module is unloaded.  This is not really necessary but its clean.
-     It allows to compare using this device WITH MTRR with another
-     method WITHOUT MTRR, without being confused by sticky MTRR settings
-     for instance.
-
-   1998-10-18   Carlo Wood      <carlo@alinoe.com>
-
-   * Forgot the '== -EBUSY' in init_module(), causing the registration of
-     character device to always fail.  You need to reboot if you tried
-     the previous version (if there is junk in `cat /proc/devices' major 107).
-
-   1998-10-18   Carlo Wood      <carlo@alinoe.com>
-
-   * struct file_operations fops_3dfx was initialized wrong:
-     `release_3dfx' was put on the place of `fsync', causing weird
-     behaviour :).
-
-   * Several warning fixes.
-
-   * setmtrr_3dfx did not always return a value, and the value was
-     ignored, effectively ignoring errors returned by mtrr_add().
-
+ * /dev/3dfx device for 2.x kernels with MTRR settings enabled.
+ *
+ * Original by Daryll Straus.
+ * Port to 2.1 kernel by Jon Taylor.
+ * setmtrr_3dfx() added by Jens Axboe.
+ * Combining 2.0 and 2.1/2.2 kernels into one rpm,
+ *   resetmtrr_3df(), and using correct pci calls for
+ *   2.1/2.2 kernels by Carlo Wood.
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software Foundation,
+ * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
+ *
  */
 
-/* Include this first as it defines things that affect the kernel headers */
+/*
+ * Include this first as it defines things that affect the kernel headers.
+ */
 #include <linux/autoconf.h>
 #include <linux/version.h>
 
@@ -125,8 +36,10 @@
 
 #define KERNEL_MIN_VER(a,b,c) (LINUX_VERSION_CODE >= KERNEL_VERSION(a,b,c))
 
-#if !KERNEL_MIN_VER (2, 1, 115)
-/* It might work with smaller kernels, but I never tested that */
+#if !KERNEL_MIN_VER(2,1,115)
+/*
+ * It might work with smaller kernels, but I never tested that.
+ */
 #error "Upgrade your kernel"
 #endif
 
@@ -155,8 +68,13 @@
 #include <linux/mm.h>
 #include <linux/errno.h>
 #include <linux/pci.h>
-#if KERNEL_MIN_VER(2,3,46) || defined(DEVFS_SUPPORT)
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,46) && \
+    LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,17) || \
+    defined(DEVFS_SUPPORT)
 #include <linux/devfs_fs_kernel.h>
+#define HAVE_DEVFS 1
+#else
+#define HAVE_DEVFS 0
 #endif
 #include <asm/segment.h>
 #include <asm/ioctl.h>
@@ -165,7 +83,7 @@
 #include <asm/processor.h>
 
 #include <asm/uaccess.h>
-#ifdef HAVE_MTRR
+#ifdef CONFIG_MTRR
 #include <asm/mtrr.h>
 #endif
 
@@ -198,6 +116,10 @@
 #define PCI_VENDOR_ID_ALLIANCE 0x1142
 #endif
 
+#ifndef PCI_VENDOR_ID_MACRONIX
+#define PCI_VENDOR_ID_MACRONIX 0x10d9
+#endif
+
 #ifndef PCI_DEVICE_ID_3DFX_VOODOO2
 #define PCI_DEVICE_ID_3DFX_VOODOO2 2
 #endif
@@ -206,6 +128,10 @@
 #define PCI_DEVICE_ID_ALLIANCE_AT3D 0x643d
 #endif
 
+#ifndef PCI_DEVICE_ID_MACRONIX_MX86251
+#define PCI_DEVICE_ID_MACRONIX_MX86251 0x8626
+#endif
+
 #ifndef PCI_DEVICE_ID_3DFX_BANSHEE
 #define PCI_DEVICE_ID_3DFX_BANSHEE 3
 #endif
@@ -221,22 +147,34 @@
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
 static char name_3dfx[] = "3dfx";
 
-#define pcibios_read_config_byte(x,y,z,w) pci_read_config_byte(y,z,w)
-#define pcibios_read_config_word(x,y,z,w) pci_read_config_word(y,z,w)
-#define pcibios_read_config_dword(x,y,z,w) pci_read_config_dword(y,z,w)
-#define pcibios_write_config_dword(x,y,z,w) pci_write_config_dword(y,z,w)
-
+static struct pci_device_id id_table_3dfx[] = {
+	{PCI_VENDOR_ID_3DFX,		PCI_DEVICE_ID_3DFX_VOODOO,
+	 PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
+	{PCI_VENDOR_ID_3DFX,		PCI_DEVICE_ID_3DFX_VOODOO2,
+	 PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
+	{PCI_VENDOR_ID_ALLIANCE,	PCI_DEVICE_ID_ALLIANCE_AT3D,
+	 PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
+	{PCI_VENDOR_ID_MACRONIX,	PCI_DEVICE_ID_MACRONIX_MX86251,
+	 PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
+	{PCI_VENDOR_ID_3DFX,		PCI_DEVICE_ID_3DFX_BANSHEE,
+	 PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
+	{PCI_VENDOR_ID_3DFX,		PCI_DEVICE_ID_3DFX_VOODOO3,
+	 PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
+	{PCI_VENDOR_ID_3DFX,		PCI_DEVICE_ID_3DFX_VOODOO4,
+	 PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
+};
 #else /* LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) */
 static struct pci_card {
 	unsigned short vendor;
 	unsigned short device;
 } pci_card_list[] = {
-	{PCI_VENDOR_ID_3DFX, 		PCI_DEVICE_ID_3DFX_VOODOO},
-	{PCI_VENDOR_ID_3DFX, 		PCI_DEVICE_ID_3DFX_VOODOO2},
-	{PCI_VENDOR_ID_ALLIANCE, 	PCI_DEVICE_ID_ALLIANCE_AT3D},
-	{PCI_VENDOR_ID_3DFX, 		PCI_DEVICE_ID_3DFX_BANSHEE},
-	{PCI_VENDOR_ID_3DFX, 		PCI_DEVICE_ID_3DFX_VOODOO3},
-	{PCI_VENDOR_ID_3DFX, 		PCI_DEVICE_ID_3DFX_VOODOO4}
+	{PCI_VENDOR_ID_3DFX,		PCI_DEVICE_ID_3DFX_VOODOO},
+	{PCI_VENDOR_ID_3DFX,		PCI_DEVICE_ID_3DFX_VOODOO2},
+	{PCI_VENDOR_ID_ALLIANCE,	PCI_DEVICE_ID_ALLIANCE_AT3D},
+	{PCI_VENDOR_ID_MACRONIX,	PCI_DEVICE_ID_MACRONIX_MX86251},
+	{PCI_VENDOR_ID_3DFX,		PCI_DEVICE_ID_3DFX_BANSHEE},
+	{PCI_VENDOR_ID_3DFX,		PCI_DEVICE_ID_3DFX_VOODOO3},
+	{PCI_VENDOR_ID_3DFX,		PCI_DEVICE_ID_3DFX_VOODOO4}
 };
 #endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) */
 
@@ -246,15 +184,22 @@ static struct pci_card {
 #define DEBUGMSG(x)
 #endif
 
-/* This macro is for accessing vma->vm_offset or vma->vm_pgoff depending
- * on kernel version */
-#if !KERNEL_MIN_VER(2, 3, 14)
+/*
+ * This macro is for accessing vma->vm_offset or vma->vm_pgoff depending
+ * on kernel version.
+ */
+#if !KERNEL_MIN_VER(2,3,14)
 #define VM_OFFSET(vma) (vma->vm_offset)
+#define VM_OFFSET_ALIGNED(vma) ((vma->vm_offset) & ~PAGE_MASK)
 #else
 #define VM_OFFSET(vma) (vma->vm_pgoff << PAGE_SHIFT)
+#define VM_OFFSET_ALIGNED(vma) ((vma->vm_pgoff) & ~PAGE_MASK)
 #endif
 
-#if KERNEL_MIN_VER(2,6,0)
+#if KERNEL_MIN_VER(2,6,10)
+#define my_remap_page_range(vma, start, ofs, len, prot) \
+	remap_pfn_range(vma, start, ofs >> PAGE_SHIFT, len, prot)
+#elif KERNEL_MIN_VER(2,6,0)
 #define my_remap_page_range(vma, start, ofs, len, prot) \
 	remap_page_range(vma, start, ofs, len, prot)
 #else
@@ -276,13 +221,9 @@ struct cardInfo_t {
 	int addr1;
 	int addr2;
 	unsigned char bus;
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
 	struct pci_dev *dev;
-#else
-	unsigned char dev;
-#endif
 	struct file *curFile;
-#ifdef HAVE_MTRR
+#ifdef CONFIG_MTRR
 	int mtrr_buf;
 	int mtrr_ctl;
 #endif
@@ -299,10 +240,12 @@ void cleanup_module(void);
 
 static cardInfo cards[MAXCARDS];
 static int numCards = 0;
+#if HAVE_DEVFS
 #if KERNEL_MIN_VER(2,3,46) && LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) \
     || defined(DEVFS_SUPPORT)
 static devfs_handle_t devfs_handle;
 #endif
+#endif
 
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
 static void findCardType(int vendor, int device)
@@ -314,7 +257,7 @@ static void findCardType(int vendor, int device)
 		pci_read_config_dword(dev, PCI_BASE_ADDRESS_1, &cards[numCards].addr1);
 		pci_read_config_dword(dev, PCI_BASE_ADDRESS_2, &cards[numCards].addr2);
 		cards[numCards].bus = dev->bus->number;
-		cards[numCards].dev = dev->devfn;
+		cards[numCards].dev = dev;
 
 		cards[numCards].addr0 &= ~0xF;
 		cards[numCards].addr1 &= ~0xF;
@@ -323,7 +266,9 @@ static void findCardType(int vendor, int device)
 		cards[numCards].curFile = 0;
 
 		DEBUGMSG(("3dfx: board vendor %d type %d located at %x/%x bus %d dev %d\n",
-		    vendor, device, cards[numCards].addr0, cards[numCards].addr1, cards[numCards].bus, cards[numCards].dev));
+			 vendor, device,
+			 cards[numCards].addr0, cards[numCards].addr1,
+			 cards[numCards].bus, cards[numCards].dev->devfn));
 
 		++numCards;
 	}
@@ -341,6 +286,20 @@ static int findCards(void)
 
 	return numCards;
 }
+#else
+static struct pci_driver driver_3dfx;
+
+static int findCards(void)
+{
+	int error;
+
+	error = pci_register_driver(&driver_3dfx);
+
+	if (error)
+		return error < 0 ? error : -error;
+	else
+		return numCards;
+}
 #endif
 
 static int open_3dfx(struct inode *inode, struct file *file)
@@ -387,18 +346,10 @@ static int mmap_3dfx(struct file *file, struct vm_area_struct *vma)
 			 VM_OFFSET(vma)));
 		return -EPERM;
 	}
-	/* This one is a special case, the macro doesn't help */
-#if !KERNEL_MIN_VER(2, 3, 14)
-	if ((vma->vm_offset) & ~PAGE_MASK) {
+	if (VM_OFFSET_ALIGNED(vma)) {
 		DEBUGMSG(("3dfx: Map request not page aligned\n"));
 		return -ENXIO;
 	}
-#else
-	if ((vma->vm_pgoff) & ~PAGE_MASK) {
-		DEBUGMSG(("3dfx: Map request not page aligned\n"));
-		return -ENXIO;
-	}
-#endif
 	len = vma->vm_end - vma->vm_start;
 	if ((len < 0) || (len > 0x2000000)) {
 		DEBUGMSG(("3dfx: Invalid mapping size requested\n"));
@@ -422,41 +373,45 @@ static int doQueryBoards(void)
 
 static int doQueryFetch(pioData *desc)
 {
-	int retval;
 	char retchar;
 	short retword;
 	int retlong;
+	int i;
 
 	if (desc->device < 0 || desc->device >= numCards)
 		return -EINVAL;
-	if ((retval = verify_area(VERIFY_WRITE, desc->value, desc->size)))
-		return retval;
+	i = desc->device;
 
 	switch (desc->port) {
 	case PCI_VENDOR_ID_LINUX:
 		if (desc->size != 2)
 			return -EINVAL;
-		copy_to_user(desc->value, &cards[desc->device].vendor, desc->size);
+		if (copy_to_user(desc->value, &cards[i].vendor, desc->size))
+			return -EFAULT;
 		return 0;
 	case PCI_DEVICE_ID_LINUX:
 		if (desc->size != 2)
 			return -EINVAL;
-		copy_to_user(desc->value, &cards[desc->device].type, desc->size);
+		if (copy_to_user(desc->value, &cards[i].type, desc->size))
+			return -EFAULT;
 		return 0;
 	case PCI_BASE_ADDRESS_0_LINUX:
 		if (desc->size != 4)
 			return -EINVAL;
-		copy_to_user(desc->value, &cards[desc->device].addr0, desc->size);
+		if (copy_to_user(desc->value, &cards[i].addr0, desc->size))
+			return -EFAULT;
 		return 0;
 	case PCI_BASE_ADDRESS_1_LINUX:
 		if (desc->size != 4)
 			return -EINVAL;
-		copy_to_user(desc->value, &cards[desc->device].addr1, desc->size);
+		if (copy_to_user(desc->value, &cards[i].addr1, desc->size))
+			return -EFAULT;
 		return 0;
 	case PCI_BASE_ADDRESS_2_LINUX:
 		if (desc->size != 4)
 			return -EINVAL;
-		copy_to_user(desc->value, &cards[desc->device].addr2, desc->size);
+		if (copy_to_user(desc->value, &cards[i].addr2, desc->size))
+			return -EFAULT;
 		return 0;
 	case SST1_PCI_SPECIAL1_LINUX:
 		if (desc->size != 4)
@@ -476,16 +431,19 @@ static int doQueryFetch(pioData *desc)
 
 	switch (desc->size) {
 	case 1:
-		pcibios_read_config_byte(cards[desc->device].bus, cards[desc->device].dev, desc->port, &retchar);
-		copy_to_user(desc->value, &retchar, 1);
+		pci_read_config_byte(cards[i].dev, desc->port, &retchar);
+		if (copy_to_user(desc->value, &retchar, 1))
+			return -EFAULT;
 		break;
 	case 2:
-		pcibios_read_config_word(cards[desc->device].bus, cards[desc->device].dev, desc->port, &retword);
-		copy_to_user(desc->value, &retword, 2);
+		pci_read_config_word(cards[i].dev, desc->port, &retword);
+		if (copy_to_user(desc->value, &retword, 2))
+			return -EFAULT;
 		break;
 	case 4:
-		pcibios_read_config_dword(cards[desc->device].bus, cards[desc->device].dev, desc->port, &retlong);
-		copy_to_user(desc->value, &retlong, 4);
+		pci_read_config_dword(cards[i].dev, desc->port, &retlong);
+		if (copy_to_user(desc->value, &retlong, 4))
+			return -EFAULT;
 		break;
 	default:
 		return -EINVAL;
@@ -505,8 +463,6 @@ static int doQueryUpdate(pioData *desc)
 
 	if (desc->device < 0 || desc->device >= numCards)
 		return -EINVAL;
-	if ((retval = verify_area(VERIFY_WRITE, desc->value, desc->size)))
-		return retval;
 
 	switch (desc->port) {
 	case PCI_COMMAND_LINUX:
@@ -533,20 +489,23 @@ static int doQueryUpdate(pioData *desc)
 		return -EINVAL;
 	}
 
-	pcibios_read_config_dword(cards[desc->device].bus, cards[desc->device].dev, desc->port & ~0x3, &retval);
+	pci_read_config_dword(cards[desc->device].dev, desc->port & ~0x3, &retval);
 	switch (desc->size) {
 	case 1:
-		copy_from_user(&retchar, desc->value, 1);
+		if (copy_from_user(&retchar, desc->value, 1))
+			return -EFAULT;
 		preval = retchar << (8 * (desc->port & 0x3));
 		mask = 0xFF << (8 * (desc->port & 0x3));
 		break;
 	case 2:
-		copy_from_user(&retword, desc->value, 2);
+		if (copy_from_user(&retword, desc->value, 2))
+			return -EFAULT;
 		preval = retword << (8 * (desc->port & 0x3));
 		mask = 0xFFFF << (8 * (desc->port & 0x3));
 		break;
 	case 4:
-		copy_from_user(&retlong, desc->value, 4);
+		if (copy_from_user(&retlong, desc->value, 4))
+			return -EFAULT;
 		preval = retlong;
 		mask = ~0;
 		break;
@@ -555,7 +514,7 @@ static int doQueryUpdate(pioData *desc)
 	}
 
 	retval = (retval & ~mask) | preval;
-	pcibios_write_config_dword(cards[desc->device].bus, cards[desc->device].dev, desc->port, retval);
+	pci_write_config_dword(cards[desc->device].dev, desc->port, retval);
 
 	return 0;
 }
@@ -563,13 +522,11 @@ static int doQueryUpdate(pioData *desc)
 static int doQuery(unsigned int cmd, unsigned long arg)
 {
 	pioData desc;
-	int retval;
 
 	if (_IOC_NR(cmd) == 2)
 		return doQueryBoards();
-	if ((retval = verify_area(VERIFY_READ, (void *)arg, sizeof(pioData))))
-		return retval;
-	copy_from_user(&desc, (void *)arg, sizeof(pioData));
+	if (copy_from_user(&desc, (void *)arg, sizeof(pioData)))
+		return -EFAULT;
 	if (_IOC_NR(cmd) == 3)
 		return doQueryFetch(&desc);
 	if (_IOC_NR(cmd) == 4)
@@ -580,13 +537,9 @@ static int doQuery(unsigned int cmd, unsigned long arg)
 
 static int doPIORead(pioData *desc)
 {
-	int retval;
 	int retchar;
 
-	if ((retval = verify_area(VERIFY_WRITE, desc->value, desc->size)))
-		return retval;
-
-	/* full SSTIO aperture is defined by: 
+	/* full SSTIO aperture is defined by:
 	 * unsigned short port = desc->port;
 	 * unsigned short base = cards[i].addr2 & ~1;
 	 * (base <= port && port <= (base + 0x108 - desc->size))
@@ -617,20 +570,18 @@ static int doPIORead(pioData *desc)
 	default:
 		return -EINVAL;
 	}
-	copy_to_user(desc->value, &retchar, desc->size);
+
+	if (copy_to_user(desc->value, &retchar, desc->size))
+		return -EFAULT;
 
 	return 0;
 }
 
 static int doPIOWrite(pioData *desc)
 {
-	int retval;
 	int retchar;
 
-	if ((retval = verify_area(VERIFY_READ, desc->value, desc->size)))
-		return retval;
-
-	/* full SSTIO aperture is defined by: 
+	/* full SSTIO aperture is defined by:
 	 * unsigned short port = desc->port;
 	 * unsigned short base = cards[i].addr2 & ~1;
 	 * (base <= port && port <= (base + 0x108 - desc->size))
@@ -646,7 +597,9 @@ static int doPIOWrite(pioData *desc)
 		return -EPERM;
 	}
 
-	copy_from_user(&retchar, desc->value, desc->size);
+	if (copy_from_user(&retchar, desc->value, desc->size))
+		return -EFAULT;
+
 	switch (desc->size) {
 	case 1:
 		outb(retchar, desc->port);
@@ -667,11 +620,9 @@ static int doPIOWrite(pioData *desc)
 static int doPIO(unsigned int cmd, unsigned long arg)
 {
 	pioData desc;
-	int retval;
 
-	if ((retval = verify_area(VERIFY_READ, (void *)arg, sizeof(pioData))))
-		return retval;
-	copy_from_user(&desc, (void *)arg, sizeof(pioData));
+	if (copy_from_user(&desc, (void *)arg, sizeof(pioData)))
+		return -EFAULT;
 	if (_IOC_DIR(cmd) == _IOC_READ)
 		return doPIORead(&desc);
 	if (_IOC_DIR(cmd) == _IOC_WRITE)
@@ -695,13 +646,15 @@ static int ioctl_3dfx(struct inode *inode, struct file *file, unsigned int cmd,
 	}
 }
 
-#ifdef HAVE_MTRR
-int setmtrr_3dfx(void)
+#ifdef CONFIG_MTRR
+static int setmtrr_3dfx(void)
 {
 	int i = 0, retval = -2;
 	unsigned char dlc;
 
-	/* First do a bios fixup if this system has a 82441FX chipset */
+	/*
+	 * First do a bios fixup if this system has a 82441FX chipset.
+	 */
 	struct pci_dev *dev = NULL;
 	dev = pci_find_device(PCI_VENDOR_ID_INTEL,
 			      PCI_DEVICE_ID_INTEL_82371SB_0, dev);
@@ -714,31 +667,39 @@ int setmtrr_3dfx(void)
 		}
 	}
 
-	/* Set up the mtrr's */
+	/*
+	 * Set up the mtrr's.
+	 */
 	if (numCards == 0)
 		return -EIO;
 	for (i = 0; i < numCards; i++) {
 		if ((cards[i].vendor != PCI_VENDOR_ID_3DFX) ||
 		    (cards[i].type > PCI_DEVICE_ID_3DFX_VOODOO3)) {
-			/* Used as flag in resetmtrr_3dfx() */
+			/*
+			 * Used as flag in resetmtrr_3dfx().
+			 */
 			cards[i].mtrr_buf = -1;
 			continue;
 		}
 		switch (cards[i].type) {
 		case PCI_DEVICE_ID_3DFX_VOODOO:
 		case PCI_DEVICE_ID_3DFX_VOODOO2:
-			/* Frame buffer to write combining */
+			/*
+			 * Frame buffer to write combining.
+			 */
 			retval = cards[i].mtrr_buf = mtrr_add(cards[i].addr0, 0x400000, MTRR_TYPE_WRCOMB, 1);
 			if (retval >= 0) {
 				retval = cards[i].mtrr_ctl = mtrr_add(cards[i].addr0, 0x1000, MTRR_TYPE_UNCACHABLE, 1);
-				if (retval<0) {
+				if (retval < 0) {
 					mtrr_del(cards[i].mtrr_buf, 0, 0);
 					cards[i].mtrr_buf = -1;
 				}
 			}
 			if (retval < 0) {
 				printk("3dfx: Could not set MTRR for Voodoo card\n");
-				/* Can still run */
+				/*
+				 * Can still run.
+				 */
 				return 0;
 			}
 			break;
@@ -747,18 +708,26 @@ int setmtrr_3dfx(void)
 			retval = cards[i].mtrr_buf = mtrr_add(cards[i].addr1, 0x1000000, MTRR_TYPE_WRCOMB, 1);
 			if (retval < 0) {
 				printk("3dfx: Could not set MTRR for Voodoo card\n");
-				/* Can still run */
+				/*
+				 * Can still run
+				 */
 				return 0;
 			}
 			cards[i].mtrr_ctl = -1;
 			break;
 		default:
-			/* We should never hit this */
+			/*
+			 * We should never hit this
+			 */
+			break;
 		}
 	}
 	if (retval == -2) {
 		DEBUGMSG(("3dfx: Could not set MTRR for this graphics card\n"));
-		retval = 0;	/* Can still run */
+		/*
+		 * Can still run
+		 */
+		retval = 0;
 	}
 #ifdef DEBUG
 	else if (retval >= 0)
@@ -769,7 +738,7 @@ int setmtrr_3dfx(void)
 	return retval;
 }
 
-int resetmtrr_3dfx(void)
+static int resetmtrr_3dfx(void)
 {
 	int i, ret, retval = 0;
 
@@ -791,15 +760,15 @@ int resetmtrr_3dfx(void)
 	return retval;
 }
 
-#endif /* HAVE_MTRR */
+#endif /* CONFIG_MTRR */
 
 static struct file_operations fops_3dfx = {
-#if KERNEL_MIN_VER(2, 4, 0)
+#if KERNEL_MIN_VER(2,4,0)
 	.owner =	THIS_MODULE,
 #endif
 	.ioctl =	ioctl_3dfx,
-	.mmap = 	mmap_3dfx,
-	.open = 	open_3dfx,
+	.mmap =		mmap_3dfx,
+	.open =		open_3dfx,
 	.release =	release_3dfx,
 };
 
@@ -830,20 +799,6 @@ static void remove_3dfx(struct pci_dev *dev)
 
 }
 
-static struct pci_device_id id_table_3dfx[] = {
-	{PCI_VENDOR_ID_3DFX,		PCI_DEVICE_ID_3DFX_VOODOO,
-	 PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
-	{PCI_VENDOR_ID_3DFX, 		PCI_DEVICE_ID_3DFX_VOODOO2,
-	 PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
-	{PCI_VENDOR_ID_ALLIANCE, 	PCI_DEVICE_ID_ALLIANCE_AT3D,
-	 PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
-	{PCI_VENDOR_ID_3DFX, 		PCI_DEVICE_ID_3DFX_BANSHEE,
-	 PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
-	{PCI_VENDOR_ID_3DFX, 		PCI_DEVICE_ID_3DFX_VOODOO3,
-	 PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
-	{PCI_VENDOR_ID_3DFX, 		PCI_DEVICE_ID_3DFX_VOODOO4,
-	 PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
-};
 MODULE_DEVICE_TABLE(pci, id_table_3dfx);
 
 static struct pci_driver driver_3dfx = {
@@ -852,11 +807,6 @@ static struct pci_driver driver_3dfx = {
 	.probe =	probe_3dfx,
 	.remove =	remove_3dfx,
 };
-
-static void findCards(void)
-{
-	pci_register_driver(&driver_3dfx);
-}
 #endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) */
 
 #ifdef MODULE
@@ -871,6 +821,7 @@ int init_module(void)
 		return ret;
 	}
 
+#if HAVE_DEVFS
 #if KERNEL_MIN_VER(2,6,0)
 	devfs_mk_cdev(MKDEV(MAJOR_3DFX, DEVICE_VOODOO),
 		      S_IFCHR | S_IROTH | S_IWOTH | S_IRGRP | S_IWGRP,
@@ -880,12 +831,17 @@ int init_module(void)
 		MAJOR_3DFX, DEVICE_VOODOO,
 		S_IFCHR | S_IROTH | S_IWOTH | S_IRGRP | S_IWGRP,
 		&fops_3dfx, NULL);
+#endif
 #endif
 
 	DEBUGMSG(("3dfx: Successfully registered device 3dfx\n"));
-	findCards();
+	ret = findCards();
+	if (ret < 0) {
+		printk("3dfx: findCards() failed, returned %d\n", ret);
+		return ret;
+	}
 
-#ifdef HAVE_MTRR
+#ifdef CONFIG_MTRR
 	ret = setmtrr_3dfx();
 	if (ret < 0) {
 		DEBUGMSG(("setmtrr_3dfx() failed, returned %d\n", ret));
@@ -903,13 +859,15 @@ void cleanup_module(void)
 {
 	DEBUGMSG(("3dfx: Entering cleanup_module\n"));
 
-#ifdef HAVE_MTRR
+#ifdef CONFIG_MTRR
 	resetmtrr_3dfx();
 #endif
+#if HAVE_DEVFS
 #if KERNEL_MIN_VER(2,6,0)
 	devfs_remove("3dfx");
 #elif KERNEL_MIN_VER(2,3,46) || defined(DEVFS_SUPPORT)
 	devfs_unregister(devfs_handle);
+#endif
 #endif
 
 	if (unregister_chrdev(MAJOR_3DFX, "3dfx")) {
@@ -925,11 +883,18 @@ void cleanup_module(void)
 
 long init_3dfx(long mem_start, long mem_end)
 {
+	int ret;
+
 	if (register_chrdev(MAJOR_3DFX, "3dfx", &fops_3dfx)) {
 		DEBUGMSG(("3dfx: Unable to register_chrdev with major %d\n", MAJOR_3DFX));
 		return 0;
 	}
-	findCards();
+
+	ret = findCards();
+	if (ret < 0) {
+		printk("3dfx: findCards() failed, returned %d\n", ret);
+		return 0;
+	}
 
 	return mem_start;
 }
@@ -940,7 +905,7 @@ MODULE_AUTHOR("Daryll Strauss et al.");
 MODULE_DESCRIPTION("3dfx glide kernel device driver");
 MODULE_LICENSE("GPL");
 #ifdef MODULE_VERSION
-MODULE_VERSION("2004.05.05");
+MODULE_VERSION("2007.02.06");
 #endif
 #endif
 
@@ -959,20 +924,5 @@ MODULE_VERSION("2004.05.05");
 #define extern
 #undef _ASM_IO_H
 #include <asm/io.h>
-/*
- * And this is copied from asm/uaccess.h
- * in order to get rid of
- * ./3dfx.o: unresolved symbol verify_area
- * when compiling this module with -g during debugging.
- */
-inline int verify_area(int type, const void * addr, unsigned long size)
-{
-	return access_ok(type,addr,size) ? 0 : -EFAULT;
-}
-
-int a_last_dummy_function(int a)
-{
-	return a;
-}
 #endif
 
